# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from dataclasses import dataclass, fields
from typing import Optional

from bl_hector.domain.administration.enumerations import Permissions as P
from bl_hector.domain.administration.repositories import Permissions
from bl_hector.domain.administration.value_objects import UserId
from bl_hector.domain.collection_management import validators
from bl_hector.domain.collection_management.entities import Book
from bl_hector.domain.collection_management.exceptions import IncorrectValue
from bl_hector.domain.collection_management.repositories import Books
from bl_hector.domain.collection_management.value_objects import (
    Author,
    Cover,
    Genre,
    Isbn,
    Title,
    Year,
)


@dataclass(frozen=True)
class Request:
    user_id: str
    isbn: str
    title: str
    year: int
    authors: list[str]
    genres: list[str]
    cover: str = ""


@dataclass
class Errors:
    isbn: Optional[IncorrectValue] = None
    title: Optional[IncorrectValue] = None
    year: Optional[IncorrectValue] = None
    authors: Optional[IncorrectValue] = None

    def __bool__(self) -> bool:
        return any([getattr(self, f.name) for f in fields(self)])


class Presenter(ABC):
    @abstractmethod
    def not_authorized(self) -> None:
        ...

    @abstractmethod
    def bad_request(self, errors: Errors) -> None:
        ...

    @abstractmethod
    def book_already_exists(self, book: Book) -> None:
        ...

    @abstractmethod
    def book_added(self, book: Book) -> None:
        ...


@dataclass(frozen=True)
class Interactor:
    presenter: Presenter
    books: Books
    permissions: Permissions

    def execute(self, request: Request) -> None:
        if not self.__is_authorized(request.user_id):
            return self.presenter.not_authorized()

        if errors := self.__validate_request(request):
            return self.presenter.bad_request(errors)

        book = Book.instanciate(
            Isbn.instanciate(request.isbn),
            Title.instanciate(request.title),
            Year.instanciate(request.year),
            [Author.instanciate(a) for a in request.authors if a],
            [Genre.instanciate(g) for g in request.genres if g],
            Cover.instanciate(request.cover) if request.cover else None,
        )

        if books := self.books.search(isbn=book.isbn):
            return self.presenter.book_already_exists(books[0])

        self.books.add(book)
        self.presenter.book_added(book)

    def __is_authorized(self, user_id: str) -> bool:
        return self.permissions.is_authorized_to(UserId(user_id), P.ADD_BOOK)

    def __validate_request(self, request: Request) -> Errors:
        return Errors(
            isbn=validators.isbn(request.isbn),
            title=validators.title(request.title),
            year=validators.year(request.year),
            authors=validators.authors(request.authors),
        )



# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from pathlib import Path
from typing import Any, Protocol

from fluent.runtime import FluentLocalization, FluentResourceLoader

DEFAULT_LOCALE = "en-GB"
LOCALES = [DEFAULT_LOCALE, "fr-FR"]
LOADER = FluentResourceLoader(str(Path(__file__).parent / "{locale}"))


class Translator(Protocol):
    def __call__(self, message_id: str, **kwargs: Any) -> str:
        ...


class DummyTranslator(Translator):
    def __call__(self, message_id: str, **kwargs: Any) -> str:
        return message_id


def localization(locale: str) -> FluentLocalization:
    if locale not in LOCALES:
        raise RuntimeError(f"Locale `{locale}` is not defined!")
    return FluentLocalization([locale] + LOCALES, ["main.ftl"], LOADER)



# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from typing import Any, Optional
from urllib.parse import urlparse

from flask import g, request, session
from sqlalchemy import create_engine
from sqlalchemy.engine import Connection

from bl_hector.domain.administration.entities import RelyingParty
from bl_hector.infrastructure import DummyPermissions
from bl_hector.infrastructure.isbnlib import BookInfoProvider
from bl_hector.infrastructure.requests import CoverProvider
from bl_hector.infrastructure.settings import WsgiSettings
from bl_hector.infrastructure.sqlalchemy.repositories import (
    Books,
    Challenges,
    Credentials,
    Users,
)
from bl_hector.interfaces import User

__SETTINGS: Optional[WsgiSettings] = None


def define_settings(settings: WsgiSettings) -> None:
    global __SETTINGS
    __SETTINGS = settings


def get_settings() -> WsgiSettings:
    if not __SETTINGS:
        raise RuntimeError("You must define the settings!")
    return __SETTINGS


def get_connection() -> Connection:
    s = get_settings()
    if "connection" not in g:
        options: dict[str, Any] = {}
        if s.DEBUG_SQL:
            options["echo"] = True
            options["echo_pool"] = "debug"

        engine = create_engine(s.DSN, **options)
        g.setdefault("connection", engine.connect())

    return g.connection  # type: ignore[no-any-return]


def teardown_connection(exception: Optional[BaseException]) -> None:
    if connection := g.pop("connection", None):
        if exception:
            connection.rollback()
        else:
            connection.commit()
        connection.close()


def get_books() -> Books:
    return Books(get_connection())


def get_users() -> Users:
    return Users(get_connection())


def get_credentials() -> Credentials:
    return Credentials(get_connection())


def get_challenges() -> Challenges:
    return Challenges(get_connection())


def get_permissions() -> DummyPermissions:
    return DummyPermissions()


def get_book_info_provider() -> BookInfoProvider:
    return BookInfoProvider()


def get_cover_provider() -> CoverProvider:
    return CoverProvider()


def get_user() -> User:
    if "user" not in g:
        if user_id := session.get("user_id", ""):
            permissions = get_permissions().for_user(user_id)
        else:
            permissions = []
        locale = g.get("locale", "")

        g.setdefault("user", User(user_id, locale, permissions))
    return g.user  # type: ignore[no-any-return]


def get_relying_party() -> RelyingParty:
    return RelyingParty(
        id=str(urlparse(request.base_url).hostname),
        name="hector",
        origin=f"http://{urlparse(request.base_url).hostname}:3000",
    )



# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from dataclasses import dataclass
from typing import Optional

from .exceptions import IncorrectValue
from .value_objects import Author, Cover, Genre, Isbn, Title, Year


@dataclass
class Book:
    isbn: Isbn
    title: Title
    year: Year
    authors: list[Author]
    genres: list[Genre]
    cover: Optional[Cover] = None

    @classmethod
    def instanciate(
        cls,
        isbn: Isbn,
        title: Title,
        year: Year,
        authors: list[Author],
        genres: list[Genre],
        cover: Optional[Cover] = None,
    ) -> "Book":
        if not authors:
            raise IncorrectValue("authors", "A book must have at least one author!")
        return Book(isbn, title, year, authors, genres, cover)



# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from dataclasses import dataclass

import bl_seth


@dataclass(frozen=True)
class WsgiSettings(bl_seth.Settings):
    SECRET_KEY: str
    """The secret key for Flask sessions.
    See: <https://flask.palletsprojects.com/en/2.3.x/quickstart/#sessions>."""

    DSN: str
    """The data source name to access the database.
    See: <https://docs.sqlalchemy.org/en/20/core/engines.html#database-urls>."""

    DEBUG_SQL: bool = False
    """To enable SqlAlchemy logging.
    See: <https://docs.sqlalchemy.org/en/20/core/engines.html#configuring-logging>."""

    PROXIED: bool = False
    """To let Flask know that it runs behind a proxy.
    See: <https://flask.palletsprojects.com/en/2.3.x/deploying/proxy_fix/>.
    """

    WEBAUTHN: bool = False
    """To enable WebAuthn authentication.
    Extra dependencies must be installed: `bl-hector[webauthn]`.
    """

    TOTP: str = ""
    """The secret key for TOTP authentication.

    Extra dependencies must be installed: `bl-hector[totp]`.
    It can be generated with `pyotp.random_base32()`.
    """


@dataclass(frozen=True)
class CliSettings(bl_seth.Settings):
    DSN: str
    """The data source name to access the database.
    See: <https://docs.sqlalchemy.org/en/20/core/engines.html#database-urls>."""

    DEBUG_SQL: bool = False
    """To enable SqlAlchemy logging.
    See: <https://docs.sqlalchemy.org/en/20/core/engines.html#configuring-logging>."""



# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from dataclasses import dataclass

from bl_hector.domain.collection_management.entities import Book
from bl_hector.domain.collection_management.exceptions import IncorrectValue
from bl_hector.domain.collection_management.repositories import Books
from bl_hector.domain.collection_management.value_objects import Isbn


@dataclass(frozen=True)
class Request:
    isbn: str


class Presenter(ABC):
    @abstractmethod
    def not_an_isbn(self, isbn: str) -> None:
        ...

    @abstractmethod
    def book_not_found(self, isbn: Isbn) -> None:
        ...

    @abstractmethod
    def see_other(self, isbn: Isbn) -> None:
        ...

    @abstractmethod
    def book(self, book: Book) -> None:
        ...


@dataclass(frozen=True)
class Interactor:
    presenter: Presenter
    books: Books

    def execute(self, request: Request) -> None:
        try:
            isbn = Isbn.instanciate(request.isbn)
        except IncorrectValue:
            return self.presenter.not_an_isbn(request.isbn)

        if str(isbn) != request.isbn:
            return self.presenter.see_other(isbn)

        if book := self.books.by_isbn(isbn):
            return self.presenter.book(book)

        self.presenter.book_not_found(isbn)



# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from enum import Enum

from bl_hector.application.use_cases import look_up_book


class ExitCodes(Enum):
    OK = 0
    USAGE = 1
    NOT_FOUND = 2
    BAD_REQUEST = 3


class WithExitCode(ABC):
    @abstractmethod
    def exit_code(self) -> int:
        ...


class LookUpBookInterface(look_up_book.Presenter, WithExitCode):
    """
    Interface required when using more than one presenter
    inside a route/command, so MyPy can do it's job!
    """



# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import logging
from typing import Any

import typer
from sqlalchemy import create_engine

from bl_hector import __version__
from bl_hector.infrastructure.settings import CliSettings
from bl_hector.infrastructure.sqlalchemy.repositories import META_DATA

from .books import cmd as books_cmd

cli = typer.Typer()
cli.add_typer(books_cmd, name="books", help="Manage books in the collection.")


@cli.callback()
def callback(quiet: bool = False, verbose: bool = False, debug: bool = False) -> None:
    level = logging.WARN
    str_fmt = "[%(asctime)s] [%(levelname)s] %(message)s"
    date_fmt = "%Y-%m-%d %H:%M:%S %z"

    if debug:
        str_fmt = str_fmt + " (%(pathname)s:%(lineno)d)"
        level = logging.DEBUG
    elif verbose:
        level = logging.INFO
    elif quiet:
        level = logging.ERROR

    logging.basicConfig(format=str_fmt, level=level, datefmt=date_fmt)


@cli.command()
def version() -> None:
    """Display application's version."""
    typer.echo(__version__)
    raise typer.Exit()


@cli.command()
def init_db(ctx: typer.Context) -> None:
    """Initialise database's schema."""
    typer.echo("Initialising schema…", nl=False)
    engine = create_engine(ctx.obj.DSN)
    META_DATA.create_all(engine)
    typer.echo(" OK!")


def build_cli(settings: CliSettings) -> Any:
    return cli(obj=settings)



# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from typing import Optional

from .entities import Book
from .value_objects import Cover, Isbn


class BookInfoProvider(ABC):
    @abstractmethod
    def look_up(self, isbn: Isbn) -> Optional[Book]:
        ...


class CoverProvider(ABC):
    @abstractmethod
    def by_isbn(self, isbn: Isbn) -> Optional[Cover]:
        ...



# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import logging
from typing import Optional

import isbnlib

from bl_hector.domain.collection_management.entities import Book
from bl_hector.domain.collection_management.services import (
    BookInfoProvider as BookInfoProviderABC,
)
from bl_hector.domain.collection_management.value_objects import (
    Author,
    Isbn,
    Title,
    Year,
)


class BookInfoProvider(BookInfoProviderABC):
    def look_up(self, isbn: Isbn) -> Optional[Book]:
        try:
            meta = isbnlib.meta(str(isbn))
            if not meta:
                return None

            return Book.instanciate(
                isbn,
                Title.instanciate(meta.get("Title", "")),
                Year.instanciate(int(meta.get("Year", ""))),
                [Author.instanciate(a) for a in meta.get("Authors", "")],
                [],
            )
        except Exception as exc:
            logging.error(str(exc))
            return None



# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from typing import Any, Optional

from werkzeug.datastructures import MultiDict

from bl_hector.application.use_cases import add_book, search_books, update_book
from bl_hector.interfaces.exceptions import BadRequest


class SearchBooks:
    def __init__(
        self, data: MultiDict[str, Any], /, *, page_size: Optional[int] = None
    ) -> None:
        self.__do_search = bool(data)
        self.request = search_books.Request(
            isbn=str(v) if (v := data.get("isbn")) else None,
            title=str(v) if (v := data.get("title")) else None,
            year=int(v) if (v := data.get("year")) else None,
            author=str(v) if (v := data.get("author")) else None,
            genre=str(v) if (v := data.get("genre")) else None,
            page_number=int(v) if (v := data.get("page")) else 1,
            page_size=page_size,
        )

    def call(self, interactor: search_books.Interactor) -> None:
        if self.__do_search:
            interactor.execute(self.request)


class AddBook:
    request: Optional[add_book.Request] = None

    def __init__(self, data: MultiDict[str, Any], user_id: str = "") -> None:
        try:
            self.request = add_book.Request(
                user_id=user_id,
                isbn=str(data.get("isbn", "")),
                title=str(data.get("title", "")),
                year=int(data.get("year", "")),
                authors=[a.strip() for a in data.get("authors", "").split(",")],
                genres=[g.strip() for g in data.get("genres", "").split(",")],
                cover=str(data.get("cover", "")),
            )
        except Exception:
            pass

    def call(self, interactor: add_book.Interactor) -> None:
        if self.request:
            interactor.execute(self.request)
        else:
            raise BadRequest()


class UpdateBook:
    request: Optional[update_book.Request] = None

    def __init__(self, isbn: str, data: MultiDict[str, Any], user_id: str = "") -> None:
        try:
            self.request = update_book.Request(
                user_id=user_id,
                isbn=isbn,
                title=str(data.get("title", "")),
                year=int(data.get("year", "")),
                authors=[a.strip() for a in data.get("authors", "").split(",")],
                genres=[g.strip() for g in data.get("genres", "").split(",")],
                cover=str(data.get("cover", "")),
            )
        except Exception:
            pass

    def call(self, interactor: update_book.Interactor) -> None:
        if self.request:
            interactor.execute(self.request)
        else:
            raise BadRequest()



# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from functools import wraps
from typing import Any, Callable

from flask import Response

from bl_hector.interfaces.to_http import HttpPresenter


def presenter_to_response(f: Callable[..., HttpPresenter]) -> Callable[[], Response]:
    @wraps(f)
    def decorated_function(*args: Any, **kwargs: Any) -> Response:
        presenter = f(*args, **kwargs)
        return Response(
            status=presenter.status_code(),
            headers=presenter.headers(),
            response=presenter.data(),
        )

    return decorated_function



# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from http import HTTPStatus as HTTP
from typing import Any

from flask import Flask, g, get_flashed_messages, request, url_for
from werkzeug.middleware.proxy_fix import ProxyFix

from bl_hector import __version__
from bl_hector.infrastructure.flask import services
from bl_hector.infrastructure.flask.aliases import blueprint as aliases
from bl_hector.infrastructure.flask.auth import blueprint as auth
from bl_hector.infrastructure.flask.books import blueprint as books
from bl_hector.infrastructure.flask.utils import presenter_to_response
from bl_hector.infrastructure.settings import WsgiSettings
from bl_hector.interfaces import exceptions, l10n
from bl_hector.interfaces.to_http.as_html import JinjaPresenter


def build_app(settings: WsgiSettings) -> Flask:
    services.define_settings(settings)

    app = Flask(__name__)

    if settings.PROXIED:
        app.wsgi_app = ProxyFix(  # type: ignore[method-assign]
            app.wsgi_app, x_for=1, x_proto=1, x_host=1, x_prefix=1
        )

    app.config.update(
        SECRET_KEY=settings.SECRET_KEY,
    )

    register_blueprints(app, settings)
    register_jinja_globals()

    app.teardown_appcontext(services.teardown_connection)

    @app.before_request
    def guess_locale() -> None:
        g.locale = (
            request.accept_languages.best_match(l10n.LOCALES) or l10n.DEFAULT_LOCALE
        )

    @app.errorhandler(exceptions.BadRequest)  # type: ignore[type-var]
    @presenter_to_response
    def handle_bad_request(e: Exception) -> Any:
        presenter = JinjaPresenter(
            "error",
            message="You submitted a bad request! This should have never happened!?",
        )
        presenter.set_status_code(HTTP.BAD_REQUEST)
        return presenter

    return app


def register_blueprints(app: Flask, settings: WsgiSettings) -> None:
    app.register_blueprint(aliases, url_prefix="/")
    app.register_blueprint(books, url_prefix="/books")

    app.register_blueprint(auth, url_prefix="/auth")
    app.auth_links = []  # type: ignore[attr-defined]
    if settings.WEBAUTHN:
        from bl_hector.infrastructure.flask.webauthn import blueprint as webauthn

        app.register_blueprint(webauthn, url_prefix="/auth/webauthn")
        app.auth_links.append(  # type: ignore[attr-defined]
            {"route": "webauthn.login", "label": "WebAuthn", "icon": "key"}
        )
    if settings.TOTP:
        from bl_hector.infrastructure.flask.totp import blueprint as totp

        app.register_blueprint(totp, url_prefix="/auth/totp")
        app.auth_links.append(  # type: ignore[attr-defined]
            {"route": "totp.login", "label": "TOTP", "icon": "clock"}
        )


def register_jinja_globals() -> None:
    from bl_hector.interfaces.to_http import as_html as presenters

    presenters.register_jinja_global("version", __version__)
    presenters.register_jinja_global("url_for", url_for)
    presenters.register_jinja_global("get_flashed_messages", get_flashed_messages)



# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from typing import Optional

from .exceptions import IncorrectValue
from .value_objects import Author, Isbn, Title, Year


def isbn(value: str) -> Optional[IncorrectValue]:
    try:
        Isbn.instanciate(value)
    except IncorrectValue as exc:
        return exc
    return None


def title(value: str) -> Optional[IncorrectValue]:
    try:
        Title.instanciate(value)
    except IncorrectValue as exc:
        return exc
    return None


def year(value: int) -> Optional[IncorrectValue]:
    try:
        Year.instanciate(value)
    except TypeError:
        # FIXME: This should be dealt with in the value object.
        return IncorrectValue("year", "TypeError")
    except IncorrectValue as exc:
        return exc
    return None


def authors(value: list[str]) -> Optional[IncorrectValue]:
    try:
        # FIXME: This domain constraint should live in the domain!
        if len([Author.instanciate(a) for a in value if a]) == 0:
            raise IncorrectValue("authors", "Cannot be empty!")
    except IncorrectValue as exc:
        return exc
    return None



# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

from bl_hector.domain.collection_management.entities import Book
from bl_hector.domain.collection_management.repositories import Books
from bl_hector.domain.collection_management.services import BookInfoProvider
from bl_hector.domain.collection_management.value_objects import Isbn


@dataclass(frozen=True)
class Request:
    path: str


class Presenter(ABC):
    @abstractmethod
    def file_does_not_exist(self, path: str) -> None:
        ...

    @abstractmethod
    def line_skipped(self, line_number: int, text: str, reason: str) -> None:
        ...

    @abstractmethod
    def book_already_exists(self, book: Book) -> None:
        ...

    @abstractmethod
    def book_info_not_found(self, isbn: Isbn) -> None:
        ...

    @abstractmethod
    def book_added(self, book: Book) -> None:
        ...


@dataclass(frozen=True)
class Interactor:
    """
    **WARNING:** This use case is not specified, for it is not trivial!
    It is also not a use case that should actually be used! It might even be removed.
    It might be better to implement it by wireing the other commands together.
    This would solve the problem of the commit of the data. For the time being
    the commit only happens once at the end of the command, even if thousands of books
    have to be imported!?
    """

    presenter: Presenter
    info_provider: BookInfoProvider
    books: Books

    def execute(self, request: Request) -> None:
        path = Path(request.path)
        if not path.is_file():
            return self.presenter.file_does_not_exist(request.path)

        for isbn in self.__extract_isbn_from_file(path):
            if book := self.books.by_isbn(isbn):
                self.presenter.book_already_exists(book)
                continue
            if not (book := self.info_provider.look_up(isbn)):
                self.presenter.book_info_not_found(isbn)
                continue
            self.books.add(book)
            self.presenter.book_added(book)

    def __extract_isbn_from_file(self, path: Path) -> list[Isbn]:
        return [
            isbn
            for number, text in enumerate(path.read_text().splitlines())
            if (isbn := self.__extract_isbn_from_line(number, text))
        ]

    def __extract_isbn_from_line(self, number: int, text: str) -> Optional[Isbn]:
        _text = text.strip()
        if "#" in _text:
            if _text.startswith("#"):
                return None
            _text = text.split("#", 1)[0]

        try:
            return Isbn.instanciate(_text)
        except Exception as exc:
            return self.presenter.line_skipped(number, text, str(exc))



# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import os

PREFIX = "HECTOR_"


def read_config_from_env() -> dict[str, str]:
    result = {}
    for k, v in os.environ.items():
        if k.startswith(PREFIX):
            result[k[len(PREFIX) :]] = v
    return result



# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from typing import Callable

from bl_hector.application.use_cases import add_book, import_books
from bl_hector.domain.collection_management.entities import Book
from bl_hector.domain.collection_management.value_objects import Isbn
from bl_hector.interfaces.l10n import DummyTranslator, Translator
from bl_hector.interfaces.to_terminal import ExitCodes, LookUpBookInterface


class LookUpBook(LookUpBookInterface):
    def __init__(
        self,
        printer: Callable[[str], None],
        /,
        *,
        translator: Translator = DummyTranslator(),
    ) -> None:
        self.__printer = printer
        self.__exit_code = ExitCodes.USAGE
        self._ = translator

    def not_an_isbn(self, isbn: str) -> None:
        self.__exit_code = ExitCodes.BAD_REQUEST
        self.__printer(self._("not-an-isbn"))

    def book_not_found(self, isbn: Isbn) -> None:
        self.__exit_code = ExitCodes.NOT_FOUND
        self.__printer(self._("book-not-found"))

    def book(self, book: Book) -> None:
        self.__exit_code = ExitCodes.OK
        self.__info(self._("book-isbn"), str(book.isbn))
        self.__info(self._("book-title"), str(book.title))
        self.__info(self._("book-year"), str(book.year))
        self.__info(self._("book-authors"), ", ".join([str(b) for b in book.authors]))
        if book.genres:
            self.__info(self._("book-genres"), ", ".join([str(g) for g in book.genres]))

    def __info(self, name: str, value: str) -> None:
        self.__printer(self._("info-line", name=name, value=value))

    def exit_code(self) -> int:
        return self.__exit_code.value


class AddBook(add_book.Presenter):
    def __init__(
        self,
        printer: Callable[[str], None],
        /,
        *,
        translator: Translator = DummyTranslator(),
    ) -> None:
        self.__printer = printer
        self.__exit_code = ExitCodes.USAGE
        self._ = translator

    def not_authorized(self) -> None:
        self.__exit_code = ExitCodes.USAGE
        self.__printer(self._("access-not-authorized"))

    def bad_request(self, errors: add_book.Errors) -> None:
        # TODO: display errors.
        self.__exit_code = ExitCodes.BAD_REQUEST
        self.__printer(self._("book-cannot-be-added"))

    def book_already_exists(self, book: Book) -> None:
        self.__exit_code = ExitCodes.BAD_REQUEST
        self.__printer(self._("book-already-exists"))

    def book_added(self, book: Book) -> None:
        self.__exit_code = ExitCodes.OK
        self.__printer(self._("book-added"))

    def exit_code(self) -> int:
        return self.__exit_code.value


class ImportBooks(import_books.Presenter):
    def __init__(
        self,
        printer: Callable[[str], None],
        # Not l10n… because it might soon be removed!
        translator: Translator = DummyTranslator(),
    ) -> None:
        self.__printer = printer
        self.__exit_code = ExitCodes.OK

    def file_does_not_exist(self, path: str) -> None:
        self.__exit_code = ExitCodes.USAGE
        self.__printer(f"File does not exist: {path}.")

    def line_skipped(self, line_number: int, text: str, reason: str) -> None:
        self.__exit_code = ExitCodes.USAGE
        self.__printer(f"Line {line_number} skipped for '{reason}' was '{text}'.")

    def book_already_exists(self, book: Book) -> None:
        self.__printer(f"Book already in the collection: {book.isbn}.")

    def book_info_not_found(self, isbn: Isbn) -> None:
        self.__printer(f"No info could be found for '{isbn}'.")

    def book_added(self, book: Book) -> None:
        self.__printer(f"Book '{book.isbn}' added: \"{book.title}\".")

    def exit_code(self) -> int:
        return self.__exit_code.value



# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import base64
import logging
from typing import Optional

import requests

from bl_hector.domain.collection_management.services import (
    CoverProvider as CoverProviderABC,
)
from bl_hector.domain.collection_management.value_objects import Cover, Isbn


class CoverProvider(CoverProviderABC):
    def by_isbn(self, isbn: Isbn) -> Optional[Cover]:
        try:
            return Cover.instanciate(
                self.__to_data_url(requests.get(self.__get_url(isbn)))
            )
        except Exception as exc:
            logging.error(str(exc))
            return None

    def __get_url(self, isbn: Isbn) -> str:
        return f"https://covers.openlibrary.org/b/isbn/{isbn}-L.jpg"

    def __to_data_url(self, response: requests.Response) -> str:
        return "data:image/jpeg;base64," + base64.b64encode(response.content).decode()



# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from typing import Type

import bl3d
from isbnlib import is_isbn13, mask

from bl_hector.domain.collection_management.exceptions import IncorrectValue


class String(bl3d.String):
    @classmethod
    def instanciate(cls: Type[bl3d.TypeString], value: str) -> bl3d.TypeString:
        try:
            return super().instanciate(value)
        except bl3d.InvalidValue as exc:
            # TODO use the `bl3d.StringXxxYyy` exceptions when available.
            #      let them pass or recast them!?
            raise IncorrectValue(cls.__name__, str(exc).split(":", 1)[1])


class Integer(bl3d.Integer):
    @classmethod
    def instanciate(cls: Type[bl3d.TypeInteger], value: int) -> bl3d.TypeInteger:
        try:
            return super().instanciate(value)
        except bl3d.InvalidValue as exc:
            # TODO use the `bl3d.NumberXxxYyy` exceptions when available.
            #      let them pass or recast them!?
            raise IncorrectValue(cls.__name__, str(exc).split(":", 1)[1])


class Isbn(String):
    """
    The International Standard Book Number (ISBN) of a book,
    for instance 978-0-7653-9277-0.
    """

    MIN = 13
    MAX = 17  # 13 numbers plus 4 `-` for masked ISBN

    @classmethod
    def instanciate(cls, value: str) -> "Isbn":
        __value = value.replace("-", "")

        super().instanciate(__value)
        if is_isbn13(__value):
            return cls(__value)

        raise IncorrectValue(cls.__name__, "This value doesn't look like an ISBN-13.")

    # TODO use the one from `bl3d` when available
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(value={self})"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Isbn):
            return NotImplemented
        # Isbn built with `__init__` might still contain `-`,
        # for instance, the ones loaded from the database
        return str(other) == str(self)

    def __str__(self) -> str:
        return str(mask(super().__str__()))


class Title(String):
    MIN = 1
    MAX = 256


class Author(String):
    MIN = 1
    MAX = 64


class Authors:
    list_of_authors: list[Author]

    def __init__(self, authors_as_list: str) -> None:
        self.list_of_authors = [Author(a.strip()) for a in authors_as_list.split(",")]

    def __str__(self) -> str:
        return ", ".join([str(a) for a in self.list_of_authors])


class Genre(String):
    MIN = 1
    MAX = 64


class Year(Integer):
    # Books published before the introduction of the ISBN system can have an ISBN!
    MIN = 1900


class Cover(String):
    MIN = 1



# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import json
from typing import Callable

from bl_hector.domain.collection_management.entities import Book
from bl_hector.domain.collection_management.value_objects import Isbn
from bl_hector.interfaces.to_terminal import ExitCodes, LookUpBookInterface


class LookUpBook(LookUpBookInterface):
    def __init__(self, printer: Callable[[str], None]) -> None:
        self.__printer = printer
        self.__exit_code = ExitCodes.USAGE

    def not_an_isbn(self, isbn: str) -> None:
        self.__exit_code = ExitCodes.BAD_REQUEST

    def book_not_found(self, isbn: Isbn) -> None:
        self.__exit_code = ExitCodes.NOT_FOUND

    def book(self, book: Book) -> None:
        self.__exit_code = ExitCodes.OK
        data = {
            "isbn": str(book.isbn),
            "title": str(book.title),
            "year": int(book.year),
            "authors": [str(b) for b in book.authors],
        }
        if book.genres:
            data["genres"] = [str(g) for g in book.genres]
        if book.cover:
            data["cover"] = str(book.cover)

        self.__printer(json.dumps(data))

    def exit_code(self) -> int:
        return self.__exit_code.value



# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from http import HTTPStatus as HTTP


class HttpPresenter(ABC):
    @abstractmethod
    def status_code(self) -> int:
        ...

    @abstractmethod
    def headers(self) -> dict[str, str]:
        ...

    @abstractmethod
    def data(self) -> str:
        ...


class Redirection(HttpPresenter):
    def __init__(self, cible: str, code_statut: HTTP = HTTP.SEE_OTHER) -> None:
        self.__cible = cible
        self.__code_statut = code_statut

    def status_code(self) -> int:
        return self.__code_statut

    def headers(self) -> dict[str, str]:
        return {"Location": self.__cible}

    def data(self) -> str:
        return ""



# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import logging
from http import HTTPStatus as HTTP
from pathlib import Path
from typing import Any, Callable, Optional

from jinja2 import Environment, FileSystemLoader, select_autoescape
from jinja2_fragments import render_block
from werkzeug.datastructures import MultiDict

from bl_hector.application.use_cases import (
    add_book,
    display_book,
    look_up_book,
    search_books,
    update_book,
)
from bl_hector.domain.collection_management import exceptions as errors
from bl_hector.domain.collection_management import validators
from bl_hector.domain.collection_management.entities import Book
from bl_hector.domain.collection_management.value_objects import Isbn
from bl_hector.interfaces import User, l10n
from bl_hector.interfaces.to_http import HttpPresenter

ENVIRONMENT = Environment(
    loader=FileSystemLoader([Path(__file__).parent / "templates"]),
    autoescape=select_autoescape(),
    extensions=["bl_hector.interfaces.utils.PatchedPyPugJSExtension"],
)


def register_jinja_global(key: str, value: Any) -> None:
    ENVIRONMENT.globals[key] = value


def url_for(*args: Any, **kwargs: Any) -> str:
    if "url_for" not in ENVIRONMENT.globals:
        raise RuntimeError("`url_for` is not declared on the environment!")
    return ENVIRONMENT.globals["url_for"](*args, **kwargs)  # type: ignore


class JinjaPresenter(HttpPresenter):
    def __init__(
        self,
        template: str,
        /,
        *,
        fragment: str = "",
        user: Optional[User] = None,
        **context: Any,
    ) -> None:
        self.__status_code = HTTP.OK
        self.__headers = {"Content-Type": "text/html; charset=UTF-8"}
        self.__template = template if template.endswith(".pug") else f"{template}.pug"
        self.__fragment = fragment
        self.__user = user
        self.__context: dict[str, Any] = {**context, "_": self._, "user": user}
        self.__error = ""

        self.__adapt_to(user)

    def __adapt_to(self, user: Optional[User] = None) -> None:
        locale = user.locale if (user and user.locale) else l10n.DEFAULT_LOCALE
        self.__localization = l10n.localization(locale)

    def _(self, message_id: str, **kwargs: Any) -> str:
        return str(self.__localization.format_value(message_id, kwargs))

    def translate_error(self, error: Optional[errors.IncorrectValue] = None) -> str:
        if not error:
            return ""

        # FIXME: Dispatching errors based on class is not very LSP!?
        if isinstance(error, errors.IncorrectValue):
            return self._("incorrect-value")
        return self._("unknown-error")

    def status_code(self) -> int:
        return int(self.__status_code)

    def headers(self) -> dict[str, str]:
        return self.__headers

    def data(self) -> str:
        return self.render()

    def set_status_code(self, status_code: HTTP) -> None:
        self.__status_code = status_code

    def set_header(self, name: str, value: str) -> None:
        self.__headers[name] = value

    def error(self, message: str, status_code: HTTP) -> None:
        self.__error = message
        self.set_status_code(status_code)

    def redirect(self, url: str, status_code: HTTP = HTTP.SEE_OTHER) -> None:
        if "Content-Type" in self.__headers:
            del self.__headers["Content-Type"]
        self.__headers["Location"] = url
        self.set_status_code(status_code)

    def not_authorized(self) -> None:
        if self.__user:
            self.set_status_code(HTTP.FORBIDDEN)
            self.__error = self._("access-forbidden")
        else:
            self.set_status_code(HTTP.UNAUTHORIZED)
            self.__error = self._("access-not-authorized")

    def render(self, **context: Any) -> str:
        if self.__error:
            return ENVIRONMENT.get_template("error.pug").render(
                **self.__context, **context, message=self.__error
            )
        if self.__fragment:
            return render_block(  # type: ignore
                ENVIRONMENT,
                self.__template,
                self.__fragment,
                **self.__context,
                **context,
            )
        return ENVIRONMENT.get_template(self.__template).render(
            **self.__context, **context
        )


class SearchBooks(JinjaPresenter, search_books.Presenter):
    PAGE_SIZE = 9

    def __init__(
        self,
        data: MultiDict[str, Any],
        /,
        *,
        fragment: str = "",
        user: Optional[User] = None,
    ) -> None:
        super().__init__("books/search", fragment=fragment, user=user)
        self.__data = data
        self.__context: dict[str, Any] = {"errors": {}}

        self.__set_page_urls()

        if data:
            # Handle the case when the search returns no book.
            self.__context["books"] = []

    def __set_page_urls(self) -> None:
        self.__context["previous_page_url"] = self.__build_url(
            {**self.__data, "page": self.__previous_page_number()}
        )
        # FIXME How to know if there are more results?!
        self.__context["next_page_url"] = self.__build_url(
            {**self.__data, "page": self.__next_page_number()}
        )

    def __build_url(self, params: dict[str, str]) -> str:
        if params.get("page", "0") == "0":
            return ""
        return url_for("books.search") + "?" + self.__join_params(params)

    def __join_params(self, params: dict[str, str]) -> str:
        return "&".join([f"{k}={v}" for k, v in params.items() if v])

    def __next_page_number(self) -> str:
        n = int(self.__data.get("page", "1"))
        return f"{n + 1}"

    def __previous_page_number(self) -> str:
        if n := int(self.__data.get("page", "0")):
            return f"{n - 1}"
        return ""

    def render(self, **context: Any) -> str:
        return super().render(**self.__context, **context, data=self.__data)

    def bad_request(self, errors: search_books.Errors) -> None:
        self.set_status_code(HTTP.OK)
        self.__context["errors"] = {
            "isbn": self.translate_error(errors.isbn),
            "title": self.translate_error(errors.title),
            "year": self.translate_error(errors.year),
            "author": self.translate_error(errors.author),
            "genre": self.translate_error(errors.genre),
        }

    def book(self, book: Book) -> None:
        if "books" not in self.__context:
            self.__context["books"] = []
        self.__context["books"].append(
            {
                "isbn": str(book.isbn),
                "title": str(book.title),
                "year": int(book.year),
                "authors": [str(a) for a in book.authors],
                "genres": [str(g) for g in book.genres],
                "cover": str(book.cover) if book.cover else "",
            }
        )


class AddBook(JinjaPresenter, add_book.Presenter):
    def __init__(
        self,
        data: MultiDict[str, Any],
        /,
        *,
        notify: Callable[[str, str], None] = lambda m, t: None,
        user: Optional[User] = None,
    ) -> None:
        super().__init__("books/add", user=user)
        self.set_status_code(HTTP.FORBIDDEN)
        self.__data = data
        self.__notify = notify
        self.__context: dict[str, Any] = {"errors": {}}

    def render(self, **context: Any) -> str:
        return super().render(**self.__context, **context, data=self.__data)

    def bad_request(self, errors: add_book.Errors) -> None:
        self.set_status_code(HTTP.OK)
        self.__context["errors"] = {
            "isbn": self.translate_error(errors.isbn),
            "title": self.translate_error(errors.title),
            "year": self.translate_error(errors.year),
            "authors": self.translate_error(errors.authors),
        }

    def book_already_exists(self, book: Book) -> None:
        self.__notify(self._("book-already-exists"), "info")
        self.redirect(url_for("books.display", isbn=str(book.isbn)))

    def book_added(self, book: Book) -> None:
        self.__notify(
            self._(
                "book-added",
                isbn=str(book.isbn),
                url=url_for("books.display", isbn=str(book.isbn)),
            ),
            "success",
        )
        self.redirect(url_for("books.add"))


class LookUpBook(JinjaPresenter, look_up_book.Presenter):
    def __init__(self, /, *, user: Optional[User] = None) -> None:
        super().__init__("books/add", fragment="form", user=user)
        self.__data: dict[str, Any] = {}
        self.__context: dict[str, Any] = {"errors": {}}

    def render(self, **context: Any) -> str:
        return super().render(**self.__context, **context, data=self.__data)

    def not_an_isbn(self, isbn: str) -> None:
        self.__data["isbn"] = isbn
        self.__context["errors"] = {"isbn": self._("not-an-isbn")}

    def unknown_error(self, message: str) -> None:
        logging.error(f"LookUpBook: {message}")
        self.__context["errors"] = {"isbn": self._("unknown-error")}

    def book_not_found(self, isbn: Isbn) -> None:
        self.__data["isbn"] = str(isbn)
        self.__context["errors"] = {"isbn": self._("unknown-isbn")}

    def book(self, book: Book) -> None:
        self.__data = {
            "isbn": str(book.isbn),
            "title": str(book.title),
            "year": int(book.year),
            "authors": ", ".join([str(a) for a in book.authors]),
            "genres": ", ".join([str(g) for g in book.genres]),
            "cover": str(book.cover),
        }


class DisplayBook(JinjaPresenter, display_book.Presenter):
    def __init__(
        self,
        /,
        *,
        notify: Callable[[str, str], None] = lambda m, t: None,
        user: Optional[User] = None,
    ) -> None:
        super().__init__("books/display", user=user)
        self.__context: dict[str, Any] = {"book": {}}
        self.__notify = notify

    def render(self, **context: Any) -> str:
        return super().render(**self.__context, **context)

    def not_an_isbn(self, isbn: str) -> None:
        self.__notify(self._("not-an-isbn"), "warning")
        self.redirect(url_for("books.search"))

    def book_not_found(self, isbn: Isbn) -> None:
        self.__notify(self._("book-does-not-exist"), "warning")
        self.redirect(url_for("books.search"))

    def see_other(self, isbn: Isbn) -> None:
        self.redirect(url_for("books.display", isbn=str(isbn)), HTTP.MOVED_PERMANENTLY)

    def book(self, book: Book) -> None:
        self.__context["book"] = {
            "isbn": str(book.isbn),
            "title": str(book.title),
            "year": int(book.year),
            "authors": ", ".join([str(a) for a in book.authors]),
            "genres": ", ".join([str(g) for g in book.genres]),
            "cover": str(book.cover) if book.cover else "",
        }


class DisplayBookToUpdate(JinjaPresenter, display_book.Presenter):
    def __init__(
        self,
        isbn: str,
        /,
        *,
        notify: Callable[[str, str], None] = lambda m, t: None,
        user: Optional[User] = None,
    ) -> None:
        super().__init__("books/update", user=user)
        self.__context: dict[str, Any] = {"isbn": isbn, "data": {}, "errors": {}}
        self.__notify = notify

    def render(self, **context: Any) -> str:
        return super().render(**self.__context, **context)

    def not_an_isbn(self, isbn: str) -> None:
        self.__notify(self._("not-an-isbn"), "warning")
        self.redirect(url_for("books.search"))

    def book_not_found(self, isbn: Isbn) -> None:
        self.__notify(self._("book-does-not-exist"), "warning")
        self.redirect(url_for("books.search"))

    def see_other(self, isbn: Isbn) -> None:
        self.redirect(url_for("books.display", isbn=str(isbn)), HTTP.MOVED_PERMANENTLY)

    def book(self, book: Book) -> None:
        self.__context["data"] = {
            "isbn": str(book.isbn),
            "title": str(book.title),
            "year": int(book.year),
            "authors": ", ".join([str(a) for a in book.authors]),
            "genres": ", ".join([str(g) for g in book.genres]),
            "cover": str(book.cover) if book.cover else "",
        }


class UpdateBook(JinjaPresenter, update_book.Presenter):
    def __init__(
        self,
        isbn: str,
        data: MultiDict[str, Any],
        /,
        *,
        notify: Callable[[str, str], None] = lambda m, t: None,
        user: Optional[User] = None,
    ) -> None:
        super().__init__("books/update", user=user)
        self.set_status_code(HTTP.FORBIDDEN)
        self.__data = data
        self.__notify = notify
        self.__context: dict[str, Any] = {"isbn": isbn, "errors": {}}

    def render(self, **context: Any) -> str:
        return super().render(**self.__context, **context, data=self.__data)

    def bad_request(self, errors: update_book.Errors) -> None:
        self.set_status_code(HTTP.OK)
        self.__context["errors"] = {
            "title": self.translate_error(errors.title),
            "year": self.translate_error(errors.year),
            "authors": self.translate_error(errors.authors),
        }

    def book_not_found(self, isbn: Isbn) -> None:
        self.__notify(self._("book-not-found"), "warning")
        self.redirect(url_for("books.search"))

    def book_updated(self, book: Book) -> None:
        self.__notify(self._("update-book-success"), "success")
        self.redirect(url_for("books.display", isbn=str(book.isbn)))


class ValidateIsbn(JinjaPresenter):
    def __init__(
        self, data: MultiDict[str, Any], /, *, user: Optional[User] = None
    ) -> None:
        super().__init__("books/search", fragment="isbn", user=user)
        self.__data = data
        self.__errors = {}

        if isbn := data.get("isbn", ""):
            errors = add_book.Errors(isbn=validators.isbn(isbn))
            self.__errors["isbn"] = self.translate_error(errors.isbn)

    def render(self, **context: Any) -> str:
        return super().render(**context, data=self.__data, errors=self.__errors)



# Hector --- A collection manager.
# Copyright © 2023 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from dataclasses import dataclass
from typing import Any

from bl_hector.domain.administration.enumerations import Permissions


@dataclass
class User:
    id: str
    locale: str
    permissions: list[Permissions]

    def __getattribute__(self, name: str) -> Any:
        PREFIX = "can_"
        if name.startswith(PREFIX):
            return self.has_permission(name[len(PREFIX) :].upper())
        return super().__getattribute__(name)

    def has_permission(self, permission: str) -> bool:
        if not self.id:
            return False
        try:
            return Permissions[permission] in self.permissions
        except KeyError:
            return False
